# -*- coding: utf-8 -*
from __future__ import division
from brightway2 import config as base_config
from brightway2 import databases, methods, mapping
from bw2data.proxies import OneDimensionalArrayProxy, \
    CompressedSparseMatrixProxy
from bw2data.utils import MAX_INT_32
from fallbacks import dicter
from scipy.sparse.linalg import factorized, spsolve
from scipy import sparse
import numpy as np
import os
try:
    import cPickle as pickle
except ImportError:
    import pickle
try:
    from bw2speedups import indexer
except ImportError:
    from fallbacks import indexer


class LCA(object):
    def __init__(self, demand, method=None, config=None):
        self.config = config or base_config
        if isinstance(demand, (basestring, tuple, list)):
            raise ValueError("Demand must be a dictionary")
        self.demand = demand
        self.method = method
        self.databases = self.get_databases()

    def get_databases(self):
        """Get list of databases for functional unit"""
        return set.union(*[set(databases[key[0]]["depends"] + [key[0]]) for key \
            in self.demand])

    def load_databases(self):
        params = np.hstack([pickle.load(open(os.path.join(
            self.config.dir, "processed", "%s.pickle" % name), "rb")
            ) for name in self.databases])
        # Technosphere
        self.tech_params = params[np.where(params['technosphere'] == True)]
        self.bio_params = params[np.where(params['technosphere'] == False)]
        self.technosphere_dict = self.build_dictionary(np.hstack((
            self.tech_params['input'], self.tech_params['output'],
            self.bio_params['output'])))
        self.add_matrix_indices(self.tech_params['input'], self.tech_params['row'],
            self.technosphere_dict)
        self.add_matrix_indices(self.tech_params['output'], self.tech_params['col'],
            self.technosphere_dict)
        # Biosphere
        self.biosphere_dict = self.build_dictionary(self.bio_params['input'])
        self.add_matrix_indices(self.bio_params['input'], self.bio_params['row'],
            self.biosphere_dict)
        self.add_matrix_indices(self.bio_params['output'], self.bio_params['col'],
            self.technosphere_dict)

    def load_method(self):
        params = pickle.load(open(os.path.join(self.config.dir, "processed",
            "%s.pickle" % methods[self.method]['abbreviation']), "rb"))
        self.add_matrix_indices(params['flow'], params['index'],
            self.biosphere_dict)
        # Eliminate references to biosphere flows that don't appear in this
        # assessment; they are masked with MAX_INT_32 values
        self.cf_params = params[np.where(params['index'] != MAX_INT_32)]

    def build_technosphere_matrix(self, vector=None):
        vector = self.tech_params['amount'] if vector is None else vector
        count = len(self.technosphere_dict)
        indices = range(count)
        # Add ones along the diagonal
        data = np.hstack((-1 * vector, np.ones((count,))))
        rows = np.hstack((self.tech_params['row'], indices))
        cols = np.hstack((self.tech_params['col'], indices))
        # coo_matrix construction is coo_matrix((values, (rows, cols)),
        # (row_count, col_count))
        self.technosphere_matrix = CompressedSparseMatrixProxy(
            sparse.coo_matrix((data, (rows, cols)), (count, count)).tocsr(),
            self.technosphere_dict, self.technosphere_dict)

    def build_biosphere_matrix(self, vector=None):
        vector = self.bio_params['amount'] if vector is None else vector
        row_count = len(self.biosphere_dict)
        col_count = len(self.technosphere_dict)
        # coo_matrix construction is coo_matrix((values, (rows, cols)),
        # (row_count, col_count))
        self.biosphere_matrix = CompressedSparseMatrixProxy(
            sparse.coo_matrix((vector, (self.bio_params['row'],
            self.bio_params['col'])), (row_count, col_count)).tocsr(),
            self.biosphere_dict, self.technosphere_dict)

    def decompose_technosphere(self):
        self.solver = factorized(self.technosphere_matrix.data.tocsc())

    def build_demand_array(self, demand=None):
        demand = demand or self.demand
        self.demand_array = OneDimensionalArrayProxy(
            np.zeros(len(self.technosphere_dict)),
            self.technosphere_dict)
        for key in demand:
            self.demand_array[mapping[key]] = demand[key]

    def build_characterization_matrix(self, vector=None):
        vector = self.cf_params['amount'] if vector is None else vector
        count = len(self.biosphere_dict)
        self.characterization_matrix = CompressedSparseMatrixProxy(
            sparse.coo_matrix((vector, (self.cf_params['index'], self.cf_params['index'])),
            (count, count)).tocsr(),
            self.biosphere_dict, self.biosphere_dict)

    def build_dictionary(self, array):
        """Build a dictionary from the sorted, unique elements of an array"""
        return dicter(array)

    def add_matrix_indices(self, array_from, array_to, mapping):
        """Map ``array_from`` keys to ``array_to`` values using ``mapping``"""
        indexer(array_from, array_to, mapping)

    def solve_linear_system(self):
        """
Master solution function for linear system a*x=b.

This is separate from 'solve_system' so that it can be easily subclassed or
replaced. Can be passed the technosphere matrix and demand vector, but is
normally can be called directly; normal parameters are assumed.

Default uses UMFpack, and stores the context so that solving subsequent times
are quick. "To most numerical analysts, matrix inversion is a sin." - Nicolas
Higham, Accuracy and Stability of Numerical Algorithms, 2002, p. 260.
        """
        if hasattr(self, "solver"):
            return self.solver(self.demand_array.data)
        else:
            return spsolve(
                self.technosphere_matrix.data,
                self.demand_array.data)

    def lci(self, factorize=False):
        """Life cycle inventory"""
        self.load_databases()
        self.build_technosphere_matrix()
        self.build_biosphere_matrix()
        self.build_demand_array()
        if factorize:
            self.decompose_technosphere()
        self.lci_calculation()

    def lci_calculation(self):
        self.supply_array = OneDimensionalArrayProxy(
            self.solve_linear_system(),
            self.technosphere_dict)
        count = len(self.technosphere_dict)
        self.inventory = CompressedSparseMatrixProxy(
            self.biosphere_matrix.data * \
            sparse.spdiags([self.supply_array.data], [0], count, count),
            self.biosphere_dict, self.technosphere_dict)

    def redo_lci(self, demand):
        """Redo LCI with same databases but different demand"""
        assert hasattr(self, "inventory"), "Must do lci first"
        self.build_demand_array(demand)
        self.lci_calculation()

    def lcia(self):
        """Life cycle impact assessment"""
        assert hasattr(self, "inventory"), "Must do lci first"
        assert self.method, "Must specify a method to perform LCIA"
        self.load_method()
        self.build_characterization_matrix()
        self.lcia_calculation()

    def lcia_calculation(self):
        self.characterized_inventory = CompressedSparseMatrixProxy(
            self.characterization_matrix.data * self.inventory.data,
            self.biosphere_dict, self.technosphere_dict)

    def redo_lcia(self, demand=None):
        assert hasattr(self, "characterized_inventory"), "Must do LCIA first"
        if demand:
            self.redo_lci(demand)
        self.lcia_calculation()

    @property
    def score(self):
        assert hasattr(self, "characterized_inventory"), "Must do LCIA first"
        return float(self.characterized_inventory.sum())

    def reverse_dict(self):
        """Construct reverse dicts from row and col indices to processes"""
        rev_mapping = dict([(v, k) for k, v in mapping.iteritems()])
        rev_tech = dict([(v, rev_mapping[k]) for k, v in \
            self.technosphere_dict.iteritems()])
        rev_bio = dict([(v, rev_mapping[k]) for k, v in \
            self.biosphere_dict.iteritems()])
        return rev_tech, rev_bio



# -*- coding: utf-8 -*
import itertools
from lca import LCA
from scipy.sparse.linalg import iterative, spsolve
from stats_toolkit.random import MCRandomNumberGenerator
import multiprocessing
import numpy as np


class MonteCarloLCA(LCA):
    def __init__(self, demand, method=None, iter_solver=iterative.cgs,
            seed=None, *args, **kwargs):
        super(MonteCarloLCA, self).__init__(demand, method=method, *args,
            **kwargs)
        self.seed = seed
        self.iter_solver = iter_solver
        self.guess = None
        self.load_databases()
        self.load_method()
        self.tech_rng = MCRandomNumberGenerator(self.tech_params, seed=seed)
        self.bio_rng = MCRandomNumberGenerator(self.bio_params, seed=seed)
        self.cf_rng = MCRandomNumberGenerator(self.cf_params, seed=seed)

    def next(self):
        self.build_technosphere_matrix(self.tech_rng.next())
        self.build_biosphere_matrix(self.bio_rng.next())
        self.build_characterization_matrix(self.cf_rng.next())

        if not hasattr(self, "demand_array"):
            self.build_demand_array()

        self.lci_calculation()
        self.lcia_calculation()

    def solve_linear_system(self):
        if not self.iter_solver or self.guess == None:
            self.guess = spsolve(
                self.technosphere_matrix.data,
                self.demand_array.data)
            return self.guess
        else:
            solution, status = self.iter_solver(
                self.technosphere_matrix.data,
                self.demand_array.data,
                x0=self.guess)
            if status != 0:
                raise
            return solution

    def iterate(self):
        # TODO: use yield or __iter__? What is most pythonic?
        self.next()
        return self.score


class ComparativeMonteCarlo(LCA):
    """First draft approach at comparative LCA"""
    def __init__(self, demands, method=None, iter_solver=iterative.cgs,
            seed=None, *args, **kwargs):
        self.demands = demands
        # Get all possibilities for database retrieval
        demand_all = demands[0].copy()
        for other in demands[1:]:
            demand_all.update(other)
        super(ComparativeMonteCarlo, self).__init__(demand_all, method)
        self.seed = seed
        self.iter_solver = iter_solver
        self.guess = None
        self.load_databases()
        self.load_method()
        self.tech_rng = MCRandomNumberGenerator(self.tech_params, seed=seed)
        self.bio_rng = MCRandomNumberGenerator(self.bio_params, seed=seed)
        self.cf_rng = MCRandomNumberGenerator(self.cf_params, seed=seed)

    def next(self):
        self.build_technosphere_matrix(self.tech_rng.next())
        self.build_biosphere_matrix(self.bio_rng.next())
        self.build_characterization_matrix(self.cf_rng.next())
        results = []
        for demand in self.demands:
            self.build_demand_array(demand)
            self.lci_calculation()
            self.lcia_calculation()
            results.append(self.score)
        return results

    def solve_linear_system(self):
        if not self.iter_solver or self.guess == None:
            self.guess = spsolve(
                self.technosphere_matrix.data,
                self.demand_array.data)
            return self.guess
        else:
            solution, status = self.iter_solver(
                self.technosphere_matrix.data,
                self.demand_array.data,
                x0=self.guess)
            if status != 0:
                raise
            return solution

    def iterate(self):
        raise NotImplemented


class ParallelMonteCarlo(object):
    """Split a Monte Carlo calculation into parallel jobs"""
    def __init__(self, demand, method, iterations=1000, chunk_size=None, cpus=None):
        self.demand = demand
        self.method = method
        if chunk_size:
            self.chunk_size = chunk_size
            self.num_jobs = iterations // chunk_size
            if iterations % self.chunk_size:
                self.num_jobs += 1
        else:
            self.num_jobs = cpus or multiprocessing.cpu_count()
            self.chunk_size = (iterations // self.num_jobs) + 1

    def calculate(self):
        pool = multiprocessing.Pool(processes=max(multiprocessing.cpu_count() - 1, 1))
        results = [pool.apply_async(single_worker, (self.demand, self.method,
            self.chunk_size)) for x in xrange(self.num_jobs)]
        pool.close()
        pool.join()  # Blocks until calculation is finished
        return list(itertools.chain(*[x.get() for x in results]))


class MultiMonteCarlo(object):
    """
This is a class for the efficient calculation of multiple demand vectors from
each Monte Carlo iteration.
    """
    def __init__(self, demands, method, iterations):
        self.demands = demands
        self.method = method
        self.iterations = iterations

    def merge_dictionaries(self, *dicts):
        r = {}
        for dic in dicts:
            for k, v in dic.iteritems():
                r.setdefault(k, []).append(v)
        return r

    def calculate(self):
        pool = multiprocessing.Pool(processes=multiprocessing.cpu_count() - 1)
        results = [pool.apply_async(multi_worker, (self.demands, self.method)
            ) for x in xrange(self.iterations)]
        pool.close()
        pool.join()  # Blocks until calculation is finished
        return self.merge_dictionaries(*[x.get() for x in results])


def single_worker(demand, method, iterations):
    # demand, method, iterations = args
    mc = MonteCarloLCA(demand=demand, method=method)
    return [mc.iterate() for x in range(iterations)]


def multi_worker(demands, method):
    lca = LCA(demands[0], method)
    lca.load_databases()
    lca.load_method()
    # Create new matrices
    lca.build_technosphere_matrix(MCRandomNumberGenerator(lca.tech_params).next())
    lca.build_biosphere_matrix(MCRandomNumberGenerator(lca.bio_params).next())
    lca.build_characterization_matrix(MCRandomNumberGenerator(lca.cf_params).next())
    lca.decompose_technosphere()
    lca.lci()
    lca.lcia()
    results = {}
    for demand in demands:
        lca.redo_lcia(demand)
        results[str(demand)] = lca.score
    return results
